/*
    Enchanced Regular User Privileges (erup)
    idea by Lukasz Wojtow <lw@wszia.edu.pl>
    code by Rafal Wijata http://www.wijata.com
*/

#define __NO_VERSION__

#include <linux/config.h>
#ifndef CONFIG_ERUP
#include <erup_conf.h>
#include <erup.h>
#else
#include <linux/erup.h>
#endif
#include <asm/uaccess.h>

/* system wide lists */
struct ug_list	*sw_uid = NULL;		/* setuid */
struct ug_list	*sw_gid = NULL;		/* setgid */
struct ch_list	*sw_chr = NULL;		/* chroot */
struct ug_list	*sw_qta = NULL;		/* quotactl */
struct ug_list	*sw_gps = NULL;		/* setgroups */
#define list_name(list)	(list == &sw_uid ? "user":\
			(list == &sw_gid ? "group" :\
			(list == &sw_qta ? "quota" :\
			(list == &sw_gps ? "groups" : "chroot") ) ) )
spinlock_t	s_lock = SPIN_LOCK_UNLOCKED;	/* lock on those lists
						    should be one for each */
extern char	infoline[];
/***************************** code starts  ************************/
#ifndef HAVESNPRINTF
/* damn, kernel <= 2.2 don't have it !
   its defined at the end of file */
int vsnprintf(char *buf, size_t size, const char *fmt, va_list args);
#endif

/* snprintf returns how much chars would be printed if there was enought
   space. It would be OK, but for zero & negative values
   it prints to buffer!
   erup_snprintf returns really written chars, no more */
static int erup_snprintf(char * buf, size_t size, const char *fmt, ...) {
va_list	args;
int	ret;
    if (size <= 0) return 0;
    va_start(args, fmt);
    ret = vsnprintf(buf, size, fmt, args);
    va_end(args);
    /* -1 because last char is overwritten in snprintf by \0 */
    return (ret >= size ? size-1 : ret);
};

/* check if pathname have symlink somewhere */
#ifndef ERUP_CHECK_SYMLINK
#define have_symlink(path, dp) (0)
#else
inline int have_symlink(const char *path, struct dentry *dp) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
struct nameidata        nd;	/* another diff, 2.2 have no path_walk */
#endif
struct dentry		*d;
int			ret;

    if (! dp) {		/* we need to lookup directory */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	path_init(path, LOOKUP_POSITIVE | LOOKUP_DIRECTORY, &nd);
	ret = path_walk(path, &nd);
	if (ret) return ret;
	d = nd.dentry;
#else
	lock_kernel();
	d = namei(path);
	ret = PTR_ERR(d);
	if (ret) { unlock_kernel(); return ret; };
#endif
    } else {		/* we got dentry, get dir from it */
	d = dp;
	ret = 0;
    };
    /* walk up to root_dir, check if it's symlink */
    for (; d && d != d->d_parent; d = d->d_parent) {
	if ( (!d->d_inode) || S_ISLNK(d->d_inode->i_mode) ) {
	    ret = -ENOTDIR;
	    goto out;
	};
    };
out:
    if (! dp)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	path_release(&nd);
#else
	{ dput(d); unlock_kernel(); };
#endif
    return ret;
};
#endif

/***************************** lists management ************************/
#if defined ERUP_SETID || defined ERUP_QUOTA || defined ERUP_SETGROUPS
/* if *id 'id' may switch to *id 'to' on list ug ?
   so both uid & gid are done here */
int is_allowed_id(id_t id, id_t to, struct ug_list *ug) {
struct ug_entry	*t_ent;
struct ug_list	*t_id;

    spin_lock(&s_lock);
    t_id = ug;
    for (t_id = ug; t_id; t_id = t_id->next) {
	if (t_id->id == id) {
	    for (t_ent = t_id->entry; t_ent; t_ent = t_ent->next) {
		if (t_ent->min <= to && to <= t_ent->max) {
		    spin_unlock(&s_lock);
		    return 1;
		};
	    };
	    break;	/* we found our id, it will not be twice */
	};
    };
    spin_unlock(&s_lock);
    return 0;
};

/* add to list *id_list entry: id may switch to (from <= x <= to) */
int erup_add_id(struct ug_list **id_list, id_t id, id_t from, id_t to) {
struct ug_entry	*t_ent;
struct ug_list	*t_id;
int		ret;

    if (from<0 || to<from || id<=0) return -EINVAL;

    ret = -ENOMEM;		/* the only error code */

    spin_lock(&s_lock);

    for (t_id = *id_list; t_id; t_id = t_id->next) if (t_id->id == id) break;

    if (! t_id) {		/* new id entry */
	t_id = (struct ug_list*)kmalloc(sizeof(struct ug_list), GFP_KERNEL);
	if (! t_id) goto out;
	t_id->id = id;
	t_id->entry = NULL;
	t_id->next = *id_list;
	*id_list = t_id;
	MOD_INC_USE_COUNT;	/* new entry - we busy */
    };

    /* check if alread on list */
    for (t_ent = t_id->entry; t_ent; t_ent = t_ent->next) {
	if (from < t_ent->min) {	/* from is below range */
	    if (to >= (t_ent->min-1)) {	/* and to is (min-1) or more */
		t_ent->min = from;
		if (to > t_ent->max) t_ent->max = to;
		goto out_ok;
	    };
	} else
	if (from <= (t_ent->max+1)) {	/* from is in range */
	    if (to <= t_ent->max) goto out_ok;	/* is within */
	    t_ent->max = to;			/* extends it */
	    goto out_ok;
	};
    };

    /* create new entry */
    t_ent = (struct ug_entry*)kmalloc(sizeof(struct ug_entry), GFP_KERNEL);
    if (! t_ent) goto out;

    t_ent->min = from;
    t_ent->max = to;

    t_ent->next = t_id->entry;
    t_id->entry = t_ent;
out_ok:
    ret = 0;

    printk(KERN_INFO DEVICENAME ": uid %d granted %d:%d-%d on %s list\n",
			    current->euid, id, from, to, list_name(id_list) );
out:
    spin_unlock(&s_lock);
    return ret;
};

/* remove ALL permissions for id swiching for id on *id_list */
int erup_del_id(struct ug_list **id_list, id_t id) {
struct ug_entry	*t_ent, *t_ent_prev;
struct ug_list	*t_id, *t_id_prev;
int		ret;

    if (id < 0) return -EINVAL;
    ret = -ENOENT;		/* the only error code */
    spin_lock(&s_lock);
again:
    for (t_id=*id_list, t_id_prev=NULL; t_id; t_id_prev=t_id, t_id=t_id->next)
	if (t_id->id==id || id==0) break;

    if (! t_id) goto out;
    
    if (! t_id_prev) {
	*id_list = t_id->next;
    } else {
	t_id_prev->next = t_id->next;
    };

    t_ent = t_id->entry;
    while (t_ent) {
	t_ent_prev = t_ent;
	t_ent = t_ent->next;
	kfree(t_ent_prev);
    };
    kfree(t_id);

    MOD_DEC_USE_COUNT;			/* we not busy */
    ret = 0;
    if (id == 0) goto again;
out:
    if (ret==0) printk(KERN_INFO DEVICENAME ": uid %d revoked %d on %s list\n",
			    current->euid, id, list_name(id_list) );
    spin_unlock(&s_lock);
    return ret;
};

#else
#define erup_add_id(id_list,id,from,to) (-ENOSYS)
#define erup_del_id(id_list,id) (-ENOSYS)
#endif	/* ERUP_SETID || ERUP_QUOTA || ERUP_SETGROUPS */


#ifndef ERUP_SETID
inline int is_allowed_uid(id_t id, id_t to) { return 0; };
inline int is_allowed_gid(id_t id, id_t to) { return 0; };
#define erup_add_uid(id,from,to)	(-ENOSYS)
#define erup_del_uid(id)		(-ENOSYS)
#define erup_add_gid(id,from,to)	(-ENOSYS)
#define erup_del_gid(id)		(-ENOSYS)

#else
inline int is_allowed_uid(id_t id, id_t to) {
    return is_allowed_id(id, to, sw_uid);
};
inline int is_allowed_gid(id_t id, id_t to) {
    return is_allowed_id(id, to, sw_gid);
};
#define erup_add_uid(id,from,to)	erup_add_id(&sw_uid,id,from,to)
#define erup_del_uid(id)		erup_del_id(&sw_uid,id)
#define erup_add_gid(id,from,to)	erup_add_id(&sw_gid,id,from,to)
#define erup_del_gid(id)		erup_del_id(&sw_gid,id)

#endif	/* ERUP_SETID */


#ifndef ERUP_CHROOT
inline int is_allowed_chr(id_t id, const char *path, struct dentry *dp) { return 0; };
#define erup_add_chr(id,path) (-ENOSYS)
#define erup_del_chr(id) (-ENOSYS)
#else
/* if uid 'id' may chroot to path or deeper ? */
inline int is_allowed_chr(id_t id, const char *path, struct dentry *dp) {
struct ch_list	*t_chr;
#ifdef ERUP_CHECK_PATH
struct ch_entry *t_ent;
#endif

    spin_lock(&s_lock);
    for (t_chr = sw_chr; t_chr; t_chr = t_chr->next) {
	if (t_chr->id == id) {
#ifdef ERUP_CHECK_PATH
	    for (t_ent = t_chr->entry; t_ent; t_ent = t_ent->next) {
#ifdef ERUP_CHECK_PATH_STRICT
		if (  (!strcmp(t_ent->path, path))
#else
		if (  (!strncmp(t_ent->path, path, t_ent->pathlen))
#endif
		    &&(!have_symlink(path, dp)) ) {
		    spin_unlock(&s_lock);
		    return 1;
		};
	    };
	    break;
#else
	    /* no path checking, user can chroot anywhere */
	    spin_unlock(&s_lock);
	    return 1;
#endif
	};
    };
    spin_unlock(&s_lock);
    return 0;
};

/* add to list entry: id may chroot() within path */
int erup_add_chr(id_t id, const char *path) {
struct ch_entry	*t_ent;
struct ch_list	*t_chr;
int		ret;

    if (path==NULL || *path!='/' || id<=0) return -EINVAL;

    ret = -ENOMEM;		/* the only error code */

    spin_lock(&s_lock);

    for (t_chr = sw_chr; t_chr; t_chr = t_chr->next) if (t_chr->id == id) break;

    if (! t_chr) {
	t_chr = (struct ch_list*)kmalloc(sizeof(struct ch_list), GFP_KERNEL);
	if (! t_chr) goto out;
	t_chr->id = id;
	t_chr->entry = NULL;
	t_chr->next = sw_chr;
	sw_chr = t_chr;
	MOD_INC_USE_COUNT;			/* we busy */
    };

#ifdef ERUP_CHECK_PATH
    /* check if alread on list */
    for (t_ent = t_chr->entry; t_ent; t_ent = t_ent->next) {
#ifdef ERUP_CHECK_PATH_STRICT
	/* only if equal */
	if (!strcmp(t_ent->path, path)) goto out_ok;
#else
	/* new is longer or equal than entry - ignore */
	if (!strncmp(t_ent->path, path, t_ent->pathlen)) goto out_ok;
	/* new is shorter than entry - replace */
	if (!strncmp(t_ent->path, path, strlen(path))) {
	    strcpy(t_ent->path, path);
	    t_ent->pathlen = strlen(t_ent->path);
	    goto out_ok;
	};
#endif
    };

    /* create new entry */
    t_ent = (struct ch_entry*)kmalloc(sizeof(struct ch_entry), GFP_KERNEL);
    if (! t_ent) goto out;

    t_ent->path = (char *)kmalloc(strlen(path) + 1, GFP_KERNEL);
    if (! t_ent->path) {
	kfree(t_ent); goto out;
    };
    strcpy(t_ent->path, path);
    t_ent->pathlen = strlen(t_ent->path);
    /* strip ending slashes, but not beggining ;) */
    while ( t_ent->pathlen > 1 && *(t_ent->path+t_ent->pathlen-1) == '/' ) {
	    t_ent->pathlen -= 1;
	    *(t_ent->path+t_ent->pathlen) = 0;
    };
    t_ent->next = t_chr->entry;
    t_chr->entry = t_ent;
out_ok:
#endif	/* ERUP_CHECK_PATH */

    ret = 0;

    printk(KERN_INFO DEVICENAME ": uid %d granted %d:%s on chroot list\n",
			    current->euid, id,          /* list_name(?) */
#ifdef ERUP_CHECK_PATH
			    t_ent->path);	/* it is safe */
#else
			    "/");
#endif
out:
    spin_unlock(&s_lock);
    return ret;
};

/* remove ALL permissions for chroot() for id */
int erup_del_chr(id_t id) {
struct ch_entry	*t_ent, *t_ent_prev;
struct ch_list	*t_chr, *t_chr_prev;
int		ret;

    if (id < 0) return -EINVAL;
    ret = -ENOENT;		/* the only error code */
    spin_lock(&s_lock);
again:
    for (t_chr=sw_chr, t_chr_prev=NULL; t_chr; t_chr_prev=t_chr, t_chr=t_chr->next)
	if (t_chr->id==id || id==0) break;

    if (! t_chr) goto out;
    
    if (! t_chr_prev) {
	sw_chr = t_chr->next;
    } else {
	t_chr_prev->next = t_chr->next;
    };

    t_ent = t_chr->entry;	/* will work no matter of ERUP_CHECK_PATH */
    while (t_ent) {
	t_ent_prev = t_ent;
	t_ent = t_ent->next;
	kfree(t_ent_prev->path);
	kfree(t_ent_prev);
    };
    kfree(t_chr);

    MOD_DEC_USE_COUNT;			/* we not busy */
    ret = 0;
    if (id == 0) goto again;
out:
    if (ret==0) printk(KERN_INFO DEVICENAME
				": uid %d revoked %d on chroot list\n",
				current->euid, id);
    spin_unlock(&s_lock);
    return ret;
};
#endif	/* ERUP_CHROOT */


#ifndef ERUP_QUOTA
inline int is_allowed_quota(id_t id, int cmd, id_t to_id) { return 0; };
#define erup_add_quota(id,read,write) (-ENOSYS)
#define erup_del_quota(id) (-ENOSYS)
#else
inline int is_allowed_quota(id_t id, int cmd, id_t to_id) {
/* cmd is after >> SUBCMDSHIFT, else will hopefully return 0 */
    /* read request */
    if (QUOTA_READ(cmd))
	return is_allowed_id(id, 1, sw_qta);
    /* write request */
    if (QUOTA_WRITE(cmd) && to_id)	/* sorry, root can't be set this way */
	return is_allowed_id(id, 2, sw_qta);
    /* admin tasks or bad cmd */
    return 0;
};
/* from == 1 -> may read
   to   == 2 -> may write */
#define erup_add_quota(id,read,write)	erup_add_id(&sw_qta, id, (read ? 1 : 2), (write ? 2 : 1))
#define erup_del_quota(id)		erup_del_id(&sw_qta, id)

#endif	/* ERUP_QUOTA */


#ifndef ERUP_SETGROUPS
inline int is_allowed_groups(id_t id, int size) { return 0; };
#define erup_add_groups(id,clear,set) (-ENOSYS)
#define erup_del_groups(id) (-ENOSYS)
#else
inline int is_allowed_groups(id_t id, int size) {
    return is_allowed_id(id, (size>0 ? 2:1), sw_gps);
};
/* from == 1 -> may clear its groups
   to   == 2 -> may set its groups */
#define erup_add_groups(id,clear,set)	erup_add_id(&sw_gps, id, (clear ? 1 : 2), (set ? 2 : 1))
#define erup_del_groups(id) 		erup_del_id(&sw_gps, id)
#endif	/* ERUP_SETGROUPS */

/************************** device functions *****************************/
/* llseek - seek on device */
loff_t erup_llseek(struct file *f, loff_t pos, int origin) {
int	ret = -ESPIPE;

    /* can only back to beginning */
    if (origin == 0 && pos == 0) {		/* SEEK_SET */
	ret = f->f_pos = pos;
    };
    return ret;
};

/* write to us */
ssize_t erup_write(struct file *f, char const *u_buf, size_t const count, loff_t *pos) {
int		ret;
id_t		id, from, to;
char		cmd;
char		*t_str;
char		*t_what;
int const	cmd_overhead = 32;
char		buf[PATH_MAX+cmd_overhead]; //should fit into kernel stack

    /* one command per one write, which means data can't be longer
       than some fixed size */
    if (count <= 0) return -EINVAL;
    if (count >= (PATH_MAX+cmd_overhead)) return -EMSGSIZE;
    ret = copy_from_user(buf, (void __user *)u_buf, count); //copy to our memory
    if (ret) return -EFAULT;
    if ( ':' != *(buf + count - 1) ) return -EBADE;
    *(buf + count - 1) = 0;	/* each command ends with `:' -> overwrite it */
    /* parse message */
    /* WARNING: it's quite big function if You intend to read it.
	It parses passed to us string, and does apropriate job */

    ret = NOERR;			/* assume no error */

    t_what = buf;			/* temporary string */

	/* +u:uid:from:to: +g:gid:from:to: +q:uid:mayread?:maywrite?:
	   +s:uid:mayclear?:mayset?:
	   +c:uid:dir: */
    if (*t_what == '+') {
	    cmd = *(++t_what);
	    if ((cmd=='u' || cmd=='g' || cmd=='q' || cmd=='s')
		&& *(t_what+1)==':') {

		    t_what += 2;
		    id = simple_strtol(t_what, &t_str, 10);
		    if (t_what==t_str || *t_str!=':') {
			ret = -EBADE;	goto parse_out;
		    };
	    
		    t_what = t_str + 1;
		    from = simple_strtol(t_what, &t_str, 10);
		    if (t_what==t_str || *t_str!=':') {
			ret = -EBADE;	goto parse_out;
		    };

		    t_what = t_str + 1;
		    to = simple_strtol(t_what, &t_str, 10);
		    if (t_what==t_str || *t_str) {
			ret = -EBADE;	goto parse_out;
		    };
	    
		    /* call the add function */
		    if (cmd == 'u') ret = erup_add_uid(id, from, to);
		    else if (cmd == 'g') ret = erup_add_gid(id, from, to);
		    else if (cmd == 'q') ret = erup_add_quota(id, from, to);
		    else if (cmd == 's') ret = erup_add_groups(id, from, to);

	    } else
	    if (cmd=='c' && *(t_what+1)==':') {

		    t_what += 2;
		    id = simple_strtol(t_what, &t_str, 10);
		    if (t_what==t_str || *t_str!=':') {
			ret = -EBADE;	goto parse_out;
		    };
	    
		    t_what = t_str + 1;

		    ret = erup_add_chr(id, t_what);

	    } else {
		    ret = -EBADE;
	    };
    } else
	/* -u:uid: -g:gid: -q:uid: -s:uid: -c:uid:
	   where if uid==0 -> remove all users */
    if (*t_what == '-') {
	    cmd = *(++t_what);
	    if ((cmd=='u' || cmd=='g' || cmd=='q' || cmd=='s' || cmd=='c')
		&& *(t_what+1)==':') {

		    t_what += 2;
		    id = simple_strtol(t_what, &t_str, 10);
		    if (t_what==t_str || *t_str) {
			ret = -EBADE;	goto parse_out;
		    };

		    if (cmd == 'u') ret = erup_del_uid(id);
		    else if (cmd == 'g') ret = erup_del_gid(id);
		    else if (cmd == 'q') ret = erup_del_quota(id);
		    else if (cmd == 's') ret = erup_del_groups(id);
		    else if (cmd == 'c') ret = erup_del_chr(id);

	    } else {
		    ret = -EBADE;
	    };
    } else {
	    ret = -EBADE;
    };

parse_out:
    if (ret) printk(KERN_WARNING DEVICENAME ": parse of '%s' is %d\n", buf, ret);
    else *pos = 0;			/* data has changed */
    /* end of parse */
    return (ret ? ret : count);	/* no err -> respond with count */
};

/* read from us */
ssize_t erup_read(struct file *f, char *u_buf, size_t const count, loff_t *pos) {
struct ug_entry	*t_ugent;
struct ug_list	*t_ug;
struct ch_entry	*t_chent;
struct ch_list	*t_chr;
char		*p_buf;
int		ret = 0; /* means no data available (yet) */
int const	count_max = 32 * 1024 - 1; /* one last byte for snprintf */
/* also it means that user may be unable to read whole config, well then let him report BUG */

    if (count <= 0) return -EINVAL;
    /* print to temp buffer, then copy to user
     * it could be printed after erup_write(), so here we would copy cached memory only
     */
    if (! (p_buf = vmalloc(count_max+1))) return -ENOMEM; /* one last byte for snprintf */
    spin_lock(&s_lock);
    ret += erup_snprintf(p_buf+ret, count_max-ret, "=%s\n", infoline);
    for (t_ug = sw_uid; t_ug; t_ug = t_ug->next)
	    for (t_ugent = t_ug->entry; t_ugent; t_ugent = t_ugent->next)
		ret += erup_snprintf(p_buf+ret, count_max-ret, "+u:%d:%d:%d\n",
			t_ug->id, t_ugent->min, t_ugent->max);
    for (t_ug = sw_gid; t_ug; t_ug = t_ug->next)
	    for (t_ugent = t_ug->entry; t_ugent; t_ugent = t_ugent->next)
		ret += erup_snprintf(p_buf+ret, count_max-ret, "+g:%d:%d:%d\n",
			t_ug->id, t_ugent->min, t_ugent->max);
    for (t_ug = sw_gps; t_ug; t_ug = t_ug->next)
		ret += erup_snprintf(p_buf+ret, count_max-ret, "+s:%d:%d:%d\n",
			t_ug->id, t_ug->entry->min, t_ug->entry->max);
    for (t_ug = sw_qta; t_ug; t_ug = t_ug->next)
		ret += erup_snprintf(p_buf+ret, count_max-ret, "+q:%d:%d:%d\n",
			t_ug->id, t_ug->entry->min, t_ug->entry->max);
    for (t_chr = sw_chr; t_chr; t_chr = t_chr->next)
#ifdef ERUP_CHECK_PATH
	for (t_chent = t_chr->entry; t_chent; t_chent = t_chent->next)
	    ret += erup_snprintf(p_buf+ret, count_max-ret, "+c:%d:%s\n",
			t_chr->id, t_chent->path);
#else
	    ret += erup_snprintf(p_buf+ret, count_max-ret, "+c:%d:%s\n", t_chr->id, "/");
#endif

    spin_unlock(&s_lock);

    /* we have to copy data and free buffer */
    if (*pos > ret) { /* *pos is no loner valid, reset to EOF */
	*pos = ret;
	ret = -ENODATA;	/* return error, let user re-read the content */
    } else { /* copy the data to user */
	ret -= *pos; /* this many bytes we may offer from that position */
	if (ret > count) ret = count; /* but if user wants less, short it */
	if ( copy_to_user((void __user *)u_buf, p_buf + *pos, ret) )
	    ret = -EFAULT; /* I don't see any reason for the partial copy failure */
    }
    
    vfree(p_buf);
    if (ret > 0) *pos += ret;	/* change position in file,
				   write to device will zero it */
    return ret;
};

/* open device */
int erup_open(struct inode *inode, struct file *file) {
	MOD_INC_USE_COUNT;
	return NOERR;
};

/* called on close - never fails */
int erup_close(struct inode *inode, struct file *file) {
	MOD_DEC_USE_COUNT;
	return NOERR;
};

/* finally define fops */
struct file_operations erup_fops = {
	llseek:		erup_llseek,
	read:		erup_read,
	write:		erup_write,
	open:		erup_open,
	release:	erup_close
};
struct proc_dir_entry	*erup_proc;

/* init function */
int erup_init(void) {
	erup_proc = create_proc_entry(DEVICENAME, S_IWUSR|S_IRUSR|S_IRGRP, &proc_root);
	if (erup_proc) {
	    erup_proc->proc_fops = &erup_fops;
	    printk(KERN_INFO "%s started\n", infoline);
	} else 
	    printk(KERN_ERR DEVICENAME ": unable register proc entry\n");
	return (erup_proc ? 0 : -EAGAIN);
};

int erup_finish(void) {
	remove_proc_entry(DEVICENAME, &proc_root);
	printk(KERN_INFO "%s stopped\n", infoline);
	return 0;
};

/* our info line */
char	infoline[] = DEVICENAME " version " ERUP_VERSION
#ifdef MODULE
	"(module)"
#endif
	 " with:"
#ifdef ERUP_SETID
	" [set*id]"
#endif
#ifdef ERUP_SETGROUPS
	" [setgroups]"
#endif
#ifdef ERUP_QUOTA
	" [quotactl]"
#endif
#ifdef ERUP_CHROOT
	" [chroot with"
#ifdef ERUP_CHECK_PATH
#ifdef ERUP_CHECK_PATH_STRICT                                                   
        " strict"
#endif
        " path check,"
#ifndef ERUP_CHECK_SYMLINK
	" no"
#endif
        " symlink check"
#else
        " no path check"
#endif
	"]"
#endif
	;

#ifndef HAVESNPRINTF
/* argh, 2.2 does not have snprintf, but also defines skip_atoi() and number()
    as static. Code below taken from 2.4 */
#define isdigit(c)	((c) >= '0' && (c) <= '9')

#define do_div(n,base) ({ \
int __res; \
__res = ((unsigned long) n) % (unsigned) base; \
n = ((unsigned long) n) / (unsigned) base; \
__res; })

static int skip_atoi(const char **s)
{
	int i=0;

	while (isdigit(**s))
		i = i*10 + *((*s)++) - '0';
	return i;
}

#define ZEROPAD	1		/* pad with zero */
#define SIGN	2		/* unsigned/signed long */
#define PLUS	4		/* show plus */
#define SPACE	8		/* space if plus */
#define LEFT	16		/* left justified */
#define SPECIAL	32		/* 0x */
#define LARGE	64		/* use 'ABCDEF' instead of 'abcdef' */

static char * number(char * buf, char * end, long long num, int base, int size, int precision, int type)
{
	char c,sign,tmp[66];
	const char *digits;
	const char small_digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
	const char large_digits[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int i;

	digits = (type & LARGE) ? large_digits : small_digits;
	if (type & LEFT)
		type &= ~ZEROPAD;
	if (base < 2 || base > 36)
		return 0;
	c = (type & ZEROPAD) ? '0' : ' ';
	sign = 0;
	if (type & SIGN) {
		if (num < 0) {
			sign = '-';
			num = -num;
			size--;
		} else if (type & PLUS) {
			sign = '+';
			size--;
		} else if (type & SPACE) {
			sign = ' ';
			size--;
		}
	}
	if (type & SPECIAL) {
		if (base == 16)
			size -= 2;
		else if (base == 8)
			size--;
	}
	i = 0;
	if (num == 0)
		tmp[i++]='0';
	else while (num != 0)
		tmp[i++] = digits[do_div(num,base)];
	if (i > precision)
		precision = i;
	size -= precision;
	if (!(type&(ZEROPAD+LEFT))) {
		while(size-->0) {
			if (buf <= end)
				*buf = ' ';
			++buf;
		}
	}
	if (sign) {
		if (buf <= end)
			*buf = sign;
		++buf;
	}
	if (type & SPECIAL) {
		if (base==8) {
			if (buf <= end)
				*buf = '0';
			++buf;
		} else if (base==16) {
			if (buf <= end)
				*buf = '0';
			++buf;
			if (buf <= end)
				*buf = digits[33];
			++buf;
		}
	}
	if (!(type & LEFT)) {
		while (size-- > 0) {
			if (buf <= end)
				*buf = c;
			++buf;
		}
	}
	while (i < precision--) {
		if (buf <= end)
			*buf = '0';
		++buf;
	}
	while (i-- > 0) {
		if (buf <= end)
			*buf = tmp[i];
		++buf;
	}
	while (size-- > 0) {
		if (buf <= end)
			*buf = ' ';
		++buf;
	}
	return buf;
}

int vsnprintf(char *buf, size_t size, const char *fmt, va_list args)
{
	int len;
	unsigned long long num;
	int i, base;
	char *str, *end, c;
	const char *s;

	int flags;		/* flags to number() */

	int field_width;	/* width of output field */
	int precision;		/* min. # of digits for integers; max
				   number of chars for from string */
	int qualifier;		/* 'h', 'l', or 'L' for integer fields */
				/* 'z' support added 23/7/1999 S.H.    */
				/* 'z' changed to 'Z' --davidm 1/25/99 */

	str = buf;
	end = buf + size - 1;

	if (end < buf - 1) {
		end = ((void *) -1);
		size = end - buf + 1;
	}

	for (; *fmt ; ++fmt) {
		if (*fmt != '%') {
			if (str <= end)
				*str = *fmt;
			++str;
			continue;
		}

		/* process flags */
		flags = 0;
		repeat:
			++fmt;		/* this also skips first '%' */
			switch (*fmt) {
				case '-': flags |= LEFT; goto repeat;
				case '+': flags |= PLUS; goto repeat;
				case ' ': flags |= SPACE; goto repeat;
				case '#': flags |= SPECIAL; goto repeat;
				case '0': flags |= ZEROPAD; goto repeat;
			}

		/* get field width */
		field_width = -1;
		if (isdigit(*fmt))
			field_width = skip_atoi(&fmt);
		else if (*fmt == '*') {
			++fmt;
			/* it's the next argument */
			field_width = va_arg(args, int);
			if (field_width < 0) {
				field_width = -field_width;
				flags |= LEFT;
			}
		}

		/* get the precision */
		precision = -1;
		if (*fmt == '.') {
			++fmt;	
			if (isdigit(*fmt))
				precision = skip_atoi(&fmt);
			else if (*fmt == '*') {
				++fmt;
				/* it's the next argument */
				precision = va_arg(args, int);
			}
			if (precision < 0)
				precision = 0;
		}

		/* get the conversion qualifier */
		qualifier = -1;
		if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L' || *fmt =='Z') {
			qualifier = *fmt;
			++fmt;
			if (qualifier == 'l' && *fmt == 'l') {
				qualifier = 'L';
				++fmt;
			}
		}

		/* default base */
		base = 10;

		switch (*fmt) {
			case 'c':
				if (!(flags & LEFT)) {
					while (--field_width > 0) {
						if (str <= end)
							*str = ' ';
						++str;
					}
				}
				c = (unsigned char) va_arg(args, int);
				if (str <= end)
					*str = c;
				++str;
				while (--field_width > 0) {
					if (str <= end)
						*str = ' ';
					++str;
				}
				continue;

			case 's':
				s = va_arg(args, char *);
				if (!s)
					s = "<NULL>";

				len = strnlen(s, precision);

				if (!(flags & LEFT)) {
					while (len < field_width--) {
						if (str <= end)
							*str = ' ';
						++str;
					}
				}
				for (i = 0; i < len; ++i) {
					if (str <= end)
						*str = *s;
					++str; ++s;
				}
				while (len < field_width--) {
					if (str <= end)
						*str = ' ';
					++str;
				}
				continue;

			case 'p':
				if (field_width == -1) {
					field_width = 2*sizeof(void *);
					flags |= ZEROPAD;
				}
				str = number(str, end,
						(unsigned long) va_arg(args, void *),
						16, field_width, precision, flags);
				continue;


			case 'n':
				/* FIXME:
				* What does C99 say about the overflow case here? */
				if (qualifier == 'l') {
					long * ip = va_arg(args, long *);
					*ip = (str - buf);
				} else if (qualifier == 'Z') {
					size_t * ip = va_arg(args, size_t *);
					*ip = (str - buf);
				} else {
					int * ip = va_arg(args, int *);
					*ip = (str - buf);
				}
				continue;

			case '%':
				if (str <= end)
					*str = '%';
				++str;
				continue;

				/* integer number formats - set up the flags and "break" */
			case 'o':
				base = 8;
				break;

			case 'X':
				flags |= LARGE;
			case 'x':
				base = 16;
				break;

			case 'd':
			case 'i':
				flags |= SIGN;
			case 'u':
				break;

			default:
				if (str <= end)
					*str = '%';
				++str;
				if (*fmt) {
					if (str <= end)
						*str = *fmt;
					++str;
				} else {
					--fmt;
				}
				continue;
		}
		if (qualifier == 'L')
			num = va_arg(args, long long);
		else if (qualifier == 'l') {
			num = va_arg(args, unsigned long);
			if (flags & SIGN)
				num = (signed long) num;
		} else if (qualifier == 'Z') {
			num = va_arg(args, size_t);
		} else if (qualifier == 'h') {
			num = (unsigned short) va_arg(args, int);
			if (flags & SIGN)
				num = (signed short) num;
		} else {
			num = va_arg(args, unsigned int);
			if (flags & SIGN)
				num = (signed int) num;
		}
		str = number(str, end, num, base,
				field_width, precision, flags);
	}
	if (str <= end)
		*str = '\0';
	else if (size > 0)
		/* don't write out a null byte if the buf size is zero */
		*end = '\0';
	/* the trailing null byte doesn't count towards the total
	* ++str;
	*/
	return str-buf;
}
#endif	/* HAVESNPRINTF */
